/*
* Copyright (c) 2019 by 3141card
* This file is released under the GPLv2.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/statvfs.h>

#include <time.h>

#include "types.h"
#include "kgen.h"
#include "device.h"
#include "util.h"
#include "fs/ufs.h"
#include "fs/fat.h"
#include "fs/ufs/ps3.h"




/***********************************************************************
* 
***********************************************************************/
void dump_sectors(ps3_context *ctx, const char *name, s64 sec_num, s64 sec_count)
{
	s64 i;
	u8 *buf = NULL;
	FILE *fd = fopen(name, "wb");
	s64 count = sec_count / 0x10000;
	s64 rest = sec_count % 0x10000;
	
	struct statvfs stat;
	statvfs(name, &stat);
	if((sec_count * SECTOR_SIZE) >= (s64)(stat.f_bsize * stat.f_bavail)) {
	  printf("Not enough free space.\n");
	  return;
	}
	
	if(count) {
		buf = malloc(0x10000 * SECTOR_SIZE);
		for(i = 0; i < count; i++) {
		  block_read(ctx, buf, 0x10000, sec_num);
		  fwrite(buf, sizeof(u8), (0x10000 * SECTOR_SIZE), fd);
		  sec_num += 0x10000;
		}
		free(buf);
  }
  
  if(rest) {
		buf = malloc(rest * SECTOR_SIZE);
		block_read(ctx, buf, rest, sec_num);
		fwrite(buf, sizeof(u8), (rest * SECTOR_SIZE), fd);
		free(buf);
	}
	
	fclose(fd);
}







/***********************************************************************
* main
***********************************************************************/
s32 main(s32 argc, char *argv[])
{
	s32 ret = -1;
	u8 *eid_root_key = {0};							// eid root key/iv
	ps3_context *ctx = NULL;            // ps3 hdd context
	struct fs *ufs2;										// superblock(UFS2)
	struct fat_bs *fat_fs;              // bootsector(FAT12/16)
	struct fat32_bs *fat32;           	// bootsector(FAT32)
	ufs_inop root_ino = ROOTINO;
	ufs_inop ino;
	
	
	// init ps3 context
	ctx = malloc(sizeof(ps3_context));
	memset(ctx, 0, sizeof(ps3_context));
	
	// load rootkey from file
	if((eid_root_key = _read_buffer((s8*)"eid_root_key", NULL)) == NULL) {		
		printf("file \"eid_root_key\" not found !\n");
		goto end;
	}
	
	// generate ata/encdec keys
	generate_ata_keys(eid_root_key, eid_root_key + 0x20, ctx->ata_k1, ctx->ata_k2);		
  generate_encdec_keys(eid_root_key, eid_root_key + 0x20, ctx->encdec_k1, ctx->encdec_k2);
	free(eid_root_key);
	
	if(argc <= 1) {
		goto end;
	}
	
	// get hdd handle and ps3 type
	if((ret = get_device_handle(ctx, argv[1])) == -1) {
		printf("%s: not a ps3 hdd or wrong eid_root_key.", argv[1]);
    goto end;
	}
	
	// get all partitions by partition table/s
	get_partitions(ctx);
	
	if(argc == 2) {
	  printf("\navailable volumes are...\n\n");
	  if(ctx->hdd0_start != 0)
	    printf(" dev_hdd0\n");
    if(ctx->hdd1_start != 0)
      printf(" dev_hdd1\n");
    if(ctx->flash_start != 0)
      printf(" dev_flash\n");
    if(ctx->flash2_start != 0)
      printf(" dev_flash2\n");
    if(ctx->flash3_start != 0)
      printf(" dev_flash3\n");
    else printf("no volumes available.\n");
	}
	
	if(argc == 3) {
		goto end;
	}
	
	// DEBUG: print sector by hdd byte offset
	if(argc == 4) {
		if(strcmp(argv[2], "print") == 0) {
			u8 *buf = malloc(SECTOR_SIZE);
			s64 sec_num = strtoll(argv[3], NULL, 16);
			block_read(ctx, buf, 1, sec_num);
			_print_buf(buf, 0, 32);
			free(buf);
		}
	}
	
	if(argc == 5) {
		if(strncmp(argv[2], "dev_", 4) == 0) {
			// if partition dev_hdd0(UFS2)...
		  if(strcmp(argv[2], "dev_hdd0") == 0) {
			  ufs2 = ufs_init(ctx);																								       // init gameOS...
			  if(ufs2 == NULL) {
			  	printf("can't open dev_hdd0!\n");
			  }
			  if(strcmp(argv[3], "dir") == 0 || strcmp(argv[3], "ls") == 0) {    		     // show dir...
			  	ufs_print_dir_list(ctx, ufs2, (u8*)argv[4], (u8*)argv[2]);    
			  	free(ufs2);
		  	}
		  	else if(strcmp(argv[3], "copy") == 0 || strcmp(argv[3], "cp") == 0) {	     // copy file/dir...
		  		ino = ufs_lookup_path(ctx, ufs2, (u8*)argv[4], 0, ROOTINO);
		  		ufs_copy_data(ctx, ufs2, root_ino, ino, argv[4], 0);
		  		free(ufs2);
		  	}
		  	else if(strcmp(argv[3], "replace") == 0) {                                 // replace file
			    ino = ufs_lookup_path(ctx, ufs2, (u8*)argv[4], 0, ROOTINO);
			    ufs_replace_data(ctx, ufs2, root_ino, ino, argv[4]);
			    free(ufs2);
		    }
			  else{
			  	free(ufs2);
		  	}
		  }
			// if partition dev_hdd1(FAT32)...
		  else if(strcmp(argv[2], "dev_hdd1") == 0) {
		  	fat32 = init_fat32(ctx, ctx->hdd1_start);
		  	if(fat32 == NULL) {
			  	printf("can't open dev_hdd1!\n");
		  	}
		  	ctx->hdd1_free = fat_how_many_free_bytes(ctx, ctx->hdd1_start, NULL, fat32);
		  	if(strcmp(argv[3], "dir") == 0 || strcmp(argv[3], "ls") == 0) {				  			// show dir...
				  fat_print_dir_list(ctx, ctx->hdd1_start, NULL, fat32, (u8*)argv[4], (u8*)argv[2], ctx->hdd1_free);
				  free(fat32);
			  }
			  else if(strcmp(argv[3], "copy") == 0 || strcmp(argv[3], "cp") == 0) {	  			// copy file/dir...
			  	fat_copy_data(ctx, ctx->hdd1_start, NULL, fat32, argv[4], 0);
			  	free(fat32);
			  }
			  else if(strcmp(argv[3], "replace") == 0) {                                 // replace file
			    fat_replace_data(ctx, ctx->hdd1_start, NULL, fat32, argv[4]);
			    free(fat32);
		    }
			  else {
		  		free(fat32);
		  	}
		  }
			// if partition dev_flash(FAT16)...
		  else if(strcmp(argv[2], "dev_flash") == 0) {	
		  	fat_fs = init_fat_old(ctx, ctx->flash_start);
		  	if(fat_fs == NULL) {
		  		printf("can't open dev_flash!\n");
			  }	
			  ctx->flash_free = fat_how_many_free_bytes(ctx, ctx->flash_start, fat_fs, NULL);
			  if(strcmp(argv[3], "dir") == 0 || strcmp(argv[3], "ls") == 0) {
			  	fat_print_dir_list(ctx, ctx->flash_start, fat_fs, NULL, (u8*)argv[4], (u8*)argv[2], ctx->flash_free);
		  		free(fat_fs);	
			  }
			  else if(strcmp(argv[3], "copy") == 0 || strcmp(argv[3], "cp") == 0) {
			  	fat_copy_data(ctx, ctx->flash_start, fat_fs, 0, argv[4], 0);
			  	free(fat_fs);
			  }
			  else if(strcmp(argv[3], "replace") == 0) {                                 // replace file
			    fat_replace_data(ctx, ctx->flash_start, fat_fs, NULL, argv[4]);
			    free(fat_fs);
		    }
			  else {
			  	free(fat_fs);
			  }
		  }
			// if partition dev_flash2(FAT16)...
		  else if(strcmp(argv[2], "dev_flash2") == 0) {
		  	fat_fs = init_fat_old(ctx, ctx->flash2_start);
		  	if(fat_fs == NULL) {
		  		printf("can't open dev_flash2!\n");
		  	}
		  	ctx->flash2_free = fat_how_many_free_bytes(ctx, ctx->flash2_start, fat_fs, NULL);
		  	if(strcmp(argv[3], "dir") == 0 || strcmp(argv[3], "ls") == 0) {
		  		fat_print_dir_list(ctx, ctx->flash2_start, fat_fs, NULL, (u8*)argv[4], (u8*)argv[2], ctx->flash2_free);
		  		free(fat_fs);
		  	}
			  else if(strcmp(argv[3], "copy") == 0 || strcmp(argv[3], "cp") == 0) {
			  	fat_copy_data(ctx, ctx->flash2_start, fat_fs, 0, argv[4], 0);
		    	free(fat_fs);
		  	}
		  	else if(strcmp(argv[3], "replace") == 0) {                                 // replace file
			    fat_replace_data(ctx, ctx->flash2_start, fat_fs, NULL, argv[4]);
			    free(fat_fs);
		    }
		  	else {
		  		free(fat_fs);
		  	}
		  }
			// if partition dev_flash3(FAT12)...
		  else if(strcmp(argv[2], "dev_flash3") == 0) {
	    	fat_fs = init_fat_old(ctx, ctx->flash3_start);
		  	if(fat_fs == NULL) {
		  		printf("can't open dev_flash3!\n");
		  	}
		  	ctx->flash3_free = fat_how_many_free_bytes(ctx, ctx->flash3_start, fat_fs, NULL);
			  if(strcmp(argv[3], "dir") == 0 || strcmp(argv[3], "ls") == 0) {
			  	fat_print_dir_list(ctx, ctx->flash3_start, fat_fs, NULL, (u8*)argv[4], (u8*)argv[2], ctx->flash3_free);
			  	free(fat_fs);	
			  }
		  	else if(strcmp(argv[3], "copy") == 0 || strcmp(argv[3], "cp") == 0) {
			  	fat_copy_data(ctx, ctx->flash3_start, fat_fs, 0, argv[4], 0);
			  	free(fat_fs);	
			  }
			  else if(strcmp(argv[3], "replace") == 0) {                                 // replace file
			    fat_replace_data(ctx, ctx->flash3_start, fat_fs, NULL, argv[4]);
			    free(fat_fs);
		    }
			  else {
			  	free(fat_fs);	
			  }
		  }
		  else {
		  	printf("no such volume!\n");
		  }
		}
	}
	
	// DEBUG: dump sector/s to file
	if(argc == 6) {
		if(strcmp(argv[2], "dump") == 0) {
			dump_sectors(ctx, argv[3], strtoll(argv[4], NULL, 16), strtoll(argv[5], NULL, 16));
		}
	}
	
end:
  if(ctx) free(ctx);
	
	return 0;
}






